"""
Predefined couplings of interest for most applications.
"""

import jax.numpy as np


def make_linear_cfun(SC, a=1.0, b=0.0):
    """Construct a linear coupling function with slope `a` and offset `b`.
    """
    def cfun(xj):
        "Compute linear coupling for pre-synaptic state variables xj."
        if xj.ndim == 1:  # no delays
            gx = SC @ xj
        elif xj.ndim == 2:  # delays
            gx = np.sum(SC * xj, axis=1)
        return a*gx + b
    return cfun


def make_diff_cfun(SC, a=1.0, b=0.0):
    """Construct a linear difference coupling."""
    nn = np.r_[:SC.shape[0]]
    diffdiag = np.diag(SC) - SC.sum(axis=1)
    SC_ = SC.at[nn,nn].set(diffdiag)
    # fix diagonal according to trick
    return make_linear_cfun(SC_, a=a, b=b)


"""
Functions for building time stepping loops.

"""

import jax
import jax.numpy as np


def make_sde(dt, dfun, gfun):
    """Use a stochastic Heun scheme to integrate `dfun` & `gfun` at a time step `dt`.

    - `dfun(x, p)` computes drift coefficients of the model
    - `gfun(x, p)` computes diffusion coefficients.

    As a shortcut, a numerical value can be passed as the `gfun` for additive SDE cases.

    This function constructs and returns two functions: step and loop.

    - `step(x, z_t, p)`: takes one step in time according to the Heun scheme. 
    - `loop(x0, zs, p)`: iteratively calls `step` for all `z`.

    In both cases, a Jax compatible parameter set `p` is provided, either an array
    or some pytree compatible structure.

    """

    sqrt_dt = np.sqrt(dt)

    # gfun is a numerical value or a function f(x,p) -> sig
    if not hasattr(gfun, '__call__'):
        sig = gfun
        gfun = lambda *_: sig

    def step(x, z_t, p):
        noise = gfun(x, p) * sqrt_dt * z_t
        d1 = dfun(x, p)
        x1 = x + dt*d1 + noise
        d2 = dfun(x1, p)
        return x + dt*0.5*(d1 + d2) + noise

    @jax.jit
    def loop(x0, zs, p):
        def op(x, z):
            x = step(x, z, p)
            return x, x
        return jax.lax.scan(op, x0, zs)[1]

    return step, loop


def make_ode(dt, dfun):
    """Use a deterministic Heun scheme to integrate `dfun` at a time step `dt`.

    - `dfun(x, p)` computes derivatives of the model

    This function constructs and returns two functions: step and loop.

    - `step(x, t, p)`: takes one step in time according to the Heun scheme. 
    - `loop(x0, ts, p)`: iteratively calls `step` for all `z`.

    In both cases, a Jax compatible parameter set `p` is provided, either an array
    or some pytree compatible structure.

    """

    def step(x, t, p):
        d1 = dfun(x, p)
        x1 = x + dt*d1
        d2 = dfun(x1, p)
        return x + dt*0.5*(d1 + d2)

    @jax.jit
    def loop(x0, ts, p):
        def op(x, t):
            x = step(x, t, p)
            return x, x
        return jax.lax.scan(op, x0, ts)[1]

    return step, loop


def make_dde(dt, nh, dfun, unroll=10):
    """Use a deterministic Heun scheme to integrate `dfun` at a time step `dt`,
    with maximum time delay (in steps) provided by `nh`. If `nh` is zero, this
    function invokes `make_ode`.

    - `dfun(xt, x, t p)` computes derivatives of the model

    where

    - `xt` is a buffer of shape (nsvar, nh + ntime)
    - `x` is the current state variable vector
    - `t` is the current time index
    - `p` is the parameter set

    The extra arguments `xt` & `t` enable time delay implementations; see
    example notebooks.

    This function constructs and returns two functions: step and loop.

    - `step(x, t, p)`: takes one step in time according to the Heun scheme. 
    - `loop(xt, ts, p)`: iteratively calls `step` for all `z`.

    In both cases, a Jax compatible parameter set `p` is provided, either an array
    or some pytree compatible structure.

    Indexing out of bounds in Jax is undefined behavior, so if `nh` is not
    large enough, or the buffer `xt` is not long enough for the delays used
    by `dfun`, the result will be silently incorrect.

    """

    if nh == 0:
        print('nh==0: using make_ode to avoid overhead')
        dfun_ = lambda x, p: dfun(x.reshape((-1, 1)), x, 0, p)
        step, loop = make_ode(dt, dfun_)
        def loop_(xt, ts, p):
            return loop(xt[:,0], ts, p).T
        return step, loop_

    def step(xt, t, p):
        x = xt[:, nh + t]
        d1 = dfun(xt, x, t, p)
        xi = x + dt*d1
        xt = xt.at[:, nh + t + 1].set(xi)
        d2 = dfun(xt, xi, t+1, p)
        nx = x + dt*0.5*(d1 + d2)
        xt = xt.at[:, nh + t + 1].set(nx)
        return xt, nx

    @jax.jit
    def loop(xt, ts, p):
        op = lambda xt, t: step(xt, t, p)
        return jax.lax.scan(op, xt, ts, unroll=unroll)[0]

    return step, loop



def make_sdde(dt, nh, dfun, gfun, unroll=10):
    "Combines semantics of make_dde & make_sde; TODO docstring"

    sqrt_dt = np.sqrt(dt)

    # TODO nh == 0: return make_sde(dt, dfun, gfun) etc

    # gfun is a numerical value or a function f(x,p) -> sig
    if not hasattr(gfun, '__call__'):
        sig = gfun
        gfun = lambda *_: sig

    def step(xt, zt, p):
        "xt is buffer, zt is (ts[i], zs[i]), p is parameters."
        t, z_t = zt
        noise = gfun(x, p) * sqrt_dt * z_t
        x = xt[:, nh + t]
        d1 = dfun(xt, x, t, p)
        xi = x + dt*d1 + noise
        xt = xt.at[:, nh + t + 1].set(xi)
        d2 = dfun(xt, xi, t+1, p)
        nx = x + dt*0.5*(d1 + d2) + noise
        xt = xt.at[:, nh + t + 1].set(nx)
        return xt, nx

    @jax.jit
    def loop(xt, zt, p):
        "xt is the buffer, zt is (ts, zs), p is parameters."
        op = lambda xt, tz: step(xt, zt, p)
        return jax.lax.scan(op, xt, zt, unroll=unroll)[0]

    return step, loop



import numpy as np
import vbjax


def test_make_dense_layers():
    """Test make_dense_layers."""
    for in_dim, extra_in, ld in [(10, 0, [10]), (10, 5, [10, 10])]:
        p, f = vbjax.make_dense_layers(in_dim, ld, extra_in=extra_in)
        x = np.random.randn(in_dim + extra_in, 1)
        y = f(p, x)
        assert y.shape == (in_dim, 1)


if __name__ == "__main__":
    test_make_dense()


import jax


def make_dense_layers(in_dim, latent_dims=[10], init_scl=0.1, extra_in=0,
               act_fn=jax.nn.leaky_relu, key=jax.random.PRNGKey(42)):
    """Make a dense neural network with the given latent layer sizes."""
    small = lambda shape: jax.random.normal(key, shape=shape)*init_scl
    
    weights = [small((latent_dims[0], in_dim + extra_in))]
    biases = [small((latent_dims[0], 1))]
    nlayers = len(latent_dims)
    
    for i in range(nlayers - 1):
        weights.append(small((latent_dims[i+1], latent_dims[i])))
        biases.append(small((latent_dims[i+1], 1)))
        
    weights.append(small((in_dim, latent_dims[-1])))
    biases.append(small((in_dim, 1)))
    
    def fwd(params, x):
        weights, biases = params
        for i in range(nlayers):
            x = act_fn(weights[i] @ x + biases[i])
        return weights[-1] @ x + biases[-1]
    
    return (weights, biases), fwd


from numpy.testing import assert_allclose
import jax.numpy as np
from vbjax import make_diff_cfun, make_linear_cfun


def test_linear_cfun():
    x = np.r_[0.0:1.0:32j]
    SC = np.r_[0.0:1.0:1024j].reshape((32, 32))
    cf = make_linear_cfun(SC, a=0.5, b=0.5)
    gx = cf(x)
    assert_allclose(gx, 0.5*SC@x+0.5, 1e-5, 1e-5)


def test_diff_cfun():
    x = np.r_[0.0:1.0:32j]
    SC = np.r_[0.0:1.0:1024j].reshape((32, 32))
    def cfun(x):
        xj = x              # receiving node i
        xi = x[:, None]     # sending node j, perhaps shape (nn, nn) due to delays
        sx = SC * (xj - xi) # weight each connection by SC
        gx = np.sum(sx, axis=1)  # sum over j
        return gx
    c1 = cfun(x)
    c2 = make_diff_cfun(SC)(x)
    assert_allclose(c1, c2, 1e-5, 1e-5)




"SHT based local coupling functions."

import numpy as np
import scipy.special as sp
try:
    import shtns
except ImportError:
    print('** shtns is not available')


# Grid functions

def make_grid_shtns(lmax, nlat, nlon, D):
    "Create shtns object and grid as in `make_grid`."
    sht = shtns.sht(lmax)
    sht.set_grid(nlat=nlat, nphi=nlon)
    theta = np.r_[:2*np.pi:1j*(nlon + 1)][:-1]
    theta += (theta[1] - theta[0]) /2
    phi = np.r_[:np.pi:1j*(nlat+1)][:-1]
    phi += (phi[1] - phi[0])/2
    gw = np.sin(phi)/phi.size/2
    return sht, phi, theta, gw


def make_lm(lmax: int):
    lm = np.array([(l,m) for m in range(lmax+1) for l in range(lmax+1) if m<=l]).T
    return lm


def make_grid(nlat, nlon):
    "Create grid for SHT, phi latitude, theta longitude."
    points, weights = np.polynomial.legendre.leggauss(nlat)
    phi = np.arccos(-points)
    gw = weights[::-1]
    theta = np.array([i*2*np.pi/nlon for i in range(nlon)])
    return phi, theta, gw


def grid_pairwise_distance(theta, phi):
    "Compute pairwise distances on grid.  Memory intensive for large grids."
    phi1 = phi.astype('f')[:,None]
    phi2 = phi1[:,None,None]
    theta1 = theta.astype('f')
    theta2 = theta1[:,None,None]
    dphi = phi2 - phi1
    dtheta = theta2 - theta1
    ds = 2*np.arcsin(np.sqrt(np.sin(dphi/2)**2 + np.cos(phi1-np.pi/2)*np.cos(phi2-np.pi/2)*np.sin(dtheta/2)**2))
    return ds


def randn_relaxed(sht):
    "For shtns sht return random spatial array captured by sht."
    return sht.synth(sht.analys(np.random.randn(*sht.spat_array().shape)))

# cortex has 0.12 m2 surface, convert to radians on circumference
C_cm = 2*np.pi*np.sqrt(0.12/(4*np.pi)) * 100
rad_to_cm = C_cm / (2 * np.pi)

# Kernel functions

def kernel_diff(D, l):
    "Compute diffusion kernel, l in shtns order."
    return D * l * (l + 1)


def kernel_dist_origin(theta, phi):
    "Compute distance to origin on grid."
    phi1 = phi[:,None]
    theta1 = theta
    theta0 = 0
    phi0 = 0
    dphi = np.abs(phi0 - phi1)
    dtheta = np.abs(theta0 - theta1)
    ds = np.arccos(np.sin(phi1-np.pi/2)*np.sin(phi0-np.pi/2) + np.cos(phi1-np.pi/2)*np.cos(phi0-np.pi/2)*np.cos(dtheta))
    return ds


def kernel_sh_normalized(sht, k):
    K = sht.analys(k)
    auc = np.sum(np.abs(K))
    return K / auc


def kernel_laplace(sht, theta, phi, size):
    "Compute spatial & spectral coefficients for Laplacian kernel."
    ds = kernel_dist_origin(theta, phi)
    return np.exp(-np.abs(ds/size))


def kernel_gaussian(sht, theta, phi, size):
    "Compute spatial & spectral coefficients for Gaussian kernel."
    ds = kernel_dist_origin(theta, phi)
    return np.sqrt(1/(size*np.pi))*np.exp(-ds**2/size)


def kernel_mexican_hat(sht, theta, phi, size):
    "Compute spatial & spectral coefficients for Mexican hat kernel."
    ds = kernel_dist_origin(theta, phi)
    k = 2/(np.sqrt(3*size)*np.pi**0.25)*(1-(ds/size)**2)*np.exp(-(ds**2/(2*size**2)))
    return k


def kernel_conv_prep(sht, k):
    "Prepares evaluated kernel for SHT convolution."
    # Driscoll & Healy 1994, Theorem 1
    K = kernel_sh_normalized(sht, k)
    im0 = np.array([sht.idx(int(_l), 0) for _l in sht.l])
    return 2*np.pi*np.sqrt(4*np.pi/(2*sht.l+1)) * K[im0]


def kernel_estimate_shtns(sht, k, x0):
    "Estimate effective kernel for kernel k with state x0 for shtns object sht."
    # uses finite differences
    f = lambda x: sht.synth(k * sht.analys(x))
    f0 = f(x0)
    dfs = np.zeros(x0.shape+x0.shape)
    for i in range(x0.shape[0]):
        for j in range(x0.shape[1]):
            x1 = x0.copy()
            x1[i,j] += 0.001
            dfs[i,j] = (f(x1) - f0) / 0.001
    gw = sht.gauss_wts()
    gw_ = np.r_[gw, gw[::-1]]
    dfs /= gw_[:,None,None,None]
    return dfs


# SHT implementation

def make_shtdiff_np(lmax, nlat, nlon, D, return_L=False, np=np):
    "Construct SHT diff implementation in plain NumPy."
    # setup grid
    lm = make_lm(lmax)
    phi, _, gw = make_grid(nlat, nlon)
    L = []
    for m in range(lmax):
        l = lm[0,lm[0]>=m]
        fwd = gw[None, :] * sp.sph_harm(m, l[:, None], 0, phi[None, :]).conjugate()
        bwd = sp.sph_harm(m, l[None, :], 0, phi[:, None])
        dll = D * l * (l + 1)
        L.append(bwd.dot(dll[:, None] * fwd))
    L = np.array(L)

    def f(x):
        X = np.fft.rfft(x, axis=1)
        X[:,:lmax] = np.einsum('abc,ca->ba',L,X[:,:lmax])
        X[:,lmax:] = 0.0
        y = np.fft.irfft(X, axis=1).real
        return y

    return (f, L) if return_L else f


def make_shtdiff(nlat, lmax=None, nlon=None, D=0.0004, return_L=False):
    "Construct SHT diff implementation with Jax."

    # imports here keeps it readable
    from numpy.testing import assert_equal
    import jax.numpy as np
    import jax

    # defaults
    nlon = nlon or 2*nlat
    lmax = lmax or nlat - 1

    # do some checking
    assert lmax < nlat
    assert nlon > nlat

    # reuse numpy impl
    _, L = make_shtdiff_np(lmax=lmax, nlat=nlat, nlon=nlon, D=D, return_L=True)
    assert_equal(0, L.imag)
    L = np.array(L.real.astype('f'))

    # create computational kernel closing over the `L` array
    def f(x):
        X = np.fft.rfft(x, axis=1)
        # can't in-place set w/ Jax so hstack padding
        X = np.hstack(
            (np.einsum('abc,ca->ba', L, X[:,:lmax]),
             np.zeros((X.shape[0], X.shape[1] - lmax), np.complex64)
            )
        )
        y = np.real(np.fft.irfft(X, axis=1))
        return y

    return (f, L) if return_L else f



import jax
import jax.numpy as np

from vbjax import make_shtdiff


def test_shtdiff():
    nlat, nlon = 32, 64
    lc = make_shtdiff(nlat=32, nlon=nlon)
    key = jax.random.PRNGKey(42)
    x0 = jax.random.normal(key, (nlat, nlon))
    x1 = lc(x0)



import jax
import jax.numpy as np

import vbjax as vb


def test_connectome_mvnorm():
    key = jax.random.PRNGKey(42)
    nc, nsc = 5, 100
    SCs = jax.random.normal(key, shape=(nsc, 32, 32))
    results = vb.make_conn_latent_mvnorm(SCs, nc=nc, return_full=True)
    u_mean, u_cov, xfm, u, s, vt, nconf = results
    assert u_mean.size == nc
    assert u_cov.shape == (nc, nc)
    assert u.shape == (nsc, nc)
    assert vt.shape == (nc, SCs.shape[1] ** 2)
    assert nconf < nsc
    *_, nconf2 = vb.make_conn_latent_mvnorm(SCs, nc=nc*2, return_full=True)
    assert nconf2 < nconf



__version__ = 'v0.0.8'

if __name__ == '__main__':
    print(__version__)
    import os, sys
    _, cmd, *args = sys.argv
    print(cmd, args)
    if cmd == 'tag':
        os.system(f'git tag {__version__}')
        os.system(f'git push -u origin {__version__}')



from .loops import make_sde, make_ode, make_dde, make_sdde
from .shtlc import make_shtdiff
from .neural_mass import JRState, JRTheta, jr_dfun, jr_default_theta
from .regmap import make_region_mapping
from .coupling import make_diff_cfun, make_linear_cfun
from .connectome import make_conn_latent_mvnorm
from .sparse import make_spmv
from .layers import make_dense_layers
from .diagnostics import shrinkage_zscore
from .embed import embed_neural_flow, embed_polynomial, embed_gradient, embed_autoregress
from ._version import __version__



import time
import numpy as np
import jax
import jax.test_util
import jax.dlpack
import jax.numpy as jnp
import jax.experimental.sparse as jsp
import scipy.sparse
import vbjax


def _test_spmv(spmv, A, n):
    nx = np.random.randn(n)
    jx = jnp.array(nx)

    # test the function itself
    jb = spmv(jx)
    nb = A @ nx
    np.testing.assert_allclose(jb, nb, 1e-6, 1e-6)

    # now its gradient
    jax.test_util.check_grads(spmv, (jx,), order=1, modes=('rev',))


def test_csr_scipy():
    n = 10
    A = scipy.sparse.random(n, n, density=0.1).tocsr()
    spmv = vbjax.make_spmv(A)
    _test_spmv(spmv, A, n)


def test_csr_scipy_symm():
    n = 10
    A = scipy.sparse.random(n, n, density=0.1).tocsr()
    A += A.T
    spmv = vbjax.make_spmv(A, is_symmetric=True)
    _test_spmv(spmv, A, n)


def _csr_to_jax_bcoo(A: scipy.sparse.csr_matrix):
    "Convert CSR format to batched COO format."
    # first convert CSR to COO
    coo = A.tocoo()
    # now convert to batched COO
    data = jax.numpy.array(coo.data)
    indices = jax.numpy.array(np.c_[coo.row, coo.col])
    shape = A.shape
    return jsp.BCOO((data, indices), shape=shape)


def bench_csr_to_bcoo():
    n = 1000
    A = scipy.sparse.random(n, n, density=0.1).tocsr()
    spmv = vbjax.make_spmv(A)
    jx = jax.numpy.r_[:n].astype('f')
    jb1 = spmv(jx)

    # now convert to bcoo
    jA = _csr_to_jax_bcoo(A)
    jspmv = jax.jit(lambda x: jA @ x)
    jb2 = jspmv(jx)

    np.testing.assert_allclose(jb1, jb2, 1e-6, 1e-6)

    for f, name in zip((spmv, jspmv), 'csr bcoo'.split(' ')):
        t0 = time.time()
        for _ in range(100):
            jb2 = f(jx)
        t1 = time.time()
        print(f'{name}: {t1 - t0:.3f} s')
    

if __name__ == '__main__':
    bench_csr_to_bcoo()



import numpy as np
import jax.dlpack
import scipy.sparse


_to_np = lambda x: np.from_dlpack(x)
_to_jax = lambda x: jax.dlpack.from_dlpack(x.__dlpack__())


def make_spmv(A, is_symmetric=False):
    """
    Make a closure for a general sparse matrix-vector multiplication.

    Parameters
    ----------
    A : scipy.sparse.csr_matrix
        Constant sparse matrix.
    is_symmetric : bool, optional, default False
        Whether matrix is symmetric.

    Returns
    -------
    spmv : function
        Function implementing spase matrix vector multiply with
        support for gradients in Jax.

    """
    AT = A.T.copy()
    @jax.custom_vjp
    def matvec(x):          return _to_jax(A @ _to_np(x))
    def matvec_tr(x):       return _to_jax(AT @ _to_np(x))
    def matvec_fwd(x):      return matvec(x), None
    def matvec_bwd(res, g): return matvec(g) if is_symmetric else matvec_tr(g),
    matvec.defvjp(matvec_fwd, matvec_bwd)
    return matvec



import numpy as np
import numpyro
import jax
import jax.numpy as jnp
from numpyro.infer import MCMC, NUTS, SVI, Trace_ELBO
import pytest

import vbjax


def setup_model():
    # setup model
    lmax, nlat, nlon = 31, 32, 64
    lc = vbjax.make_shtdiff(lmax=lmax, nlat=nlat, nlon=nlon)
    x0 = jax.random.normal(jax.random.PRNGKey(42), (nlat, nlon))
    x1 = lc(x0)
    ts = jnp.r_[:100]*0.01
    _, run = vbjax.make_ode(0.01, lambda x, k: -x + k*lc(x))
    k = 1.0
    xt = run(x0, ts, k)

    # parameter to recover
    x0 = x0.at[:15,:15].set(3.0)
    xt = run(x0, ts, k)

    # bayesian model
    import numpyro.distributions as dist
    def logp(xt=None):
        x0h = numpyro.sample('x0h', dist.Normal(jnp.zeros((nlat, nlon)), 1))
        xth_mu = run(x0h, ts, k)
        numpyro.sample('xth', dist.Normal(xth_mu, 1), obs=xt)
    return x0, xt, logp


@pytest.mark.slow
def test_sht_ode_nuts():
    x0, xt, logp = setup_model()
    nuts_kernel = NUTS(logp)
    mcmc = MCMC(nuts_kernel, num_warmup=500, num_samples=500)
    rng_key = jax.random.PRNGKey(0)
    mcmc.run(rng_key, xt=xt)
    x0h = mcmc.get_samples()['x0h']

    # check effective sample size
    ess = numpyro.diagnostics.effective_sample_size(x0h.reshape((1, 500, -1)))
    assert ess.min() > 100

    # check sbc measures
    shrinkage, zscore = vbjax.shrinkage_zscore(x0, x0h, 1)
    assert shrinkage.min() > 0.7
    assert zscore.max() < 1.5



import collections
import jax.numpy as np


JRTheta = collections.namedtuple(
    typename='JRTheta',
    field_names='A B a b v0 nu_max r J a_1 a_2 a_3 a_4 mu I'.split(' '))

jr_default_theta = JRTheta(
    A=3.25, B=22.0, a=0.1, b=0.05, v0=5.52, nu_max=0.0025, 
    r=0.56, J=135.0, a_1=1.0, a_2=0.8, a_3=0.25, a_4=0.25, mu=0.22, I=0.0)

JRState = collections.namedtuple(
    typename='JRState',
    field_names='y0 y1 y2 y3 y4 y5'.split(' '))
    
def jr_dfun(ys, c, p):
    y0, y1, y2, y3, y4, y5 = ys

    sigm_y1_y2 = 2.0 * p.nu_max / (1.0 + np.exp(p.r * (p.v0 - (y1 - y2))))
    sigm_y0_1  = 2.0 * p.nu_max / (1.0 + np.exp(p.r * (p.v0 - (p.a_1 * p.J * y0))))
    sigm_y0_3  = 2.0 * p.nu_max / (1.0 + np.exp(p.r * (p.v0 - (p.a_3 * p.J * y0))))

    return np.array([y3,
        y4,
        y5,
        p.A * p.a * sigm_y1_y2 - 2.0 * p.a * y3 - p.a ** 2 * y0,
        p.A * p.a * (p.mu + p.a_2 * p.J * sigm_y0_1 + c)
            - 2.0 * p.a * y4 - p.a ** 2 * y1,
        p.B * p.b * (p.a_4 * p.J * sigm_y0_3) - 2.0 * p.b * y5 - p.b ** 2 * y2,
                     ])


BVEPTheta = collections.namedtuple(
    typename='BVEPTheta',
    field_names='tau0 I1 eta'
)

bvep_default_theta = BVEPTheta(
    tau0=10.0, I1=3.1, eta=-3.5
)

def bvep_dfun(ys, c, p: BVEPTheta):
    z, x = ys
    x2 = x*x
    dx = 1 - x*x2 - 2*x2 - z + p.I1
    dz = (1/p.tau0)*(4*(x - p.eta) - z - c)
    return np.array([dx, dz])


# Montbrio-Pazo-Roxin
MPRTheta = collections.namedtuple(
    typename='MPRTheta',
    field_names='tau I Delta J eta cr cv'.split(' '))

mpr_default_theta = MPRTheta(
    tau=1.0,
    I=0.0,
    Delta=1.0,
    J=15.0,
    eta=-5.0,
    cr=1.0,
    cv=0.0
)

MPRState = collections.namedtuple(
    typename='MPRState',
    field_names='r V'.split(' '))

mpr_default_state = MPRState(r=0.0, V=-2.0)

def mpr_dfun(ys, c, p):
    r, V = ys

    I_c = p.cr * c[0] + p.cv * c[1]

    return np.array([
        (1 / p.tau) * (p.Delta / (np.pi * p.tau) + 2 * r * V),
        (1 / p.tau) * (V ** 2 + p.eta + p.J * p.tau *
         r + p.I + I_c - (np.pi ** 2) * (r ** 2) * (p.tau ** 2))
    ])



import jax
import jax.numpy as np

from vbjax import make_sde, make_ode, make_dde


def test_sde():
    f = lambda x,_: -x
    g = lambda *_: 1e-2
    dt = 0.1
    _, run = make_sde(dt, f, g)

    key = jax.random.PRNGKey(42)
    zs = jax.random.normal(key, (100, 32))

    xs = run(zs[0]+1, zs, None)
    assert xs.shape == zs.shape


def test_ode():
    f = lambda x,_: -x
    dt = 0.1
    _, run = make_ode(dt, f)

    x0 = np.r_[:32].astype('f')
    xs = run(x0, np.r_[:64], None)
    assert xs.shape == (64, 32)


def test_dde():
    def dfun(xt, x, t, p):
        xd = xt[0, t-100]
        dx = x - x**3/3 + p*xt[0, t-100]
        return dx

    _, loop = make_dde(0.1, 100, dfun)
    xt0 = np.ones((1, 200))
    xt1 = loop(xt0, np.r_[:100], 0.2)
    assert xt1.shape == (1, 200)



"""
These are tests of our understanding of how Jax works.  If these break
then probably things elsewhere will also break.

"""

import numpy
import jax
import jax.dlpack
import jax.test_util


def test_dlpack_numpy():
    x = numpy.random.randn(10)
    jx = jax.dlpack.from_dlpack(x.__dlpack__())
    nx = numpy.from_dlpack(jx)
    numpy.testing.assert_allclose(x, nx)

def test_custom_vjp_simple():

    @jax.custom_vjp
    def foo(a, b):
        return 2*a+b, 3*a+0.5*b

    def foo_fwd(a, b):
        return foo(a, b), None

    def foo_bwd(res, g):
        g_1, g_2 = g
        return (2*g[0]+3*g[1], g[0] + 0.5*g[1])

    foo.defvjp(foo_fwd, foo_bwd)

    def bar(a, b):
        c, d = foo(a, b)
        return 2*c + 3*d

    jax.grad(bar, [0,1])(3.,4.)

    jax.test_util.check_grads(bar, (3.0, 4.0), order=1, modes=('rev',))


if __name__ == '__main__':
    test_dlpack_numpy()
    test_custom_vjp_simple()



"""
Utilities for connectomes.

"""

import jax.numpy as np
import scipy.stats


def make_conn_latent_mvnorm(SCs, nc=10, return_full=False):
    """
    Make a latent multivariate normal distribution over connectomes.
    
    Parameters
    ----------
    SCs : (nconn, n, n) array
        Array of connectomes in a given parcellation.
    nc : int, optional
        Number of components to use for the latent space.
    return_full: bool, optional
        Whether or not to return extra information on the SVD.
    
    Returns
    -------
    u_mean : (nc, ), array_like
        Mean of distribution in latent space.
    u_cov : (nc, nc), array_like
        Covariance of distribution in latent space.
    xfm : function
        Maps latent vector to full connectome.
    u_cov : (nc, nc) array
        Covariance of the multivariate normal. Returned if return_full=True.
    u : (nconn, nconn) array
        Left singular vectors corresponding to connectomes embedded.
        Returned if return_full=True.
    s : (nconn) array
        Singular values. Returned if return_full=True.
    vt : (nconn, n*n) array
        Right singular vectors. Returned if return_full=True.
    nconf : int
        Number of confusions induced by dimensionality reduction.
        Returned if return_full=True.

    """
    nconn, nn, _ = SCs.shape
    u, s, vt = np.linalg.svd(SCs.reshape((nconn, -1)), full_matrices=False)
    u, s, vt = u[:,:nc], s[:nc], vt[:nc]
    rweights = (u @ np.diag(s) @ vt).reshape(SCs.shape)
    sse = np.sum(np.square(rweights[:,None] - SCs).reshape((nconn, nconn, -1)), axis=2)
    # TODO allow nc to be a float specifying percentage of permissible confusions
    nconf = (np.argmin(sse, axis=1) != np.r_[:nconn]).sum()
    u_mean = np.mean(u, axis=0)
    u_cov = np.cov(u.T)
    diag_s = np.diag(s)
    def xfm(new_u):
        return (new_u @ diag_s @ vt).reshape(new_u.shape[:-1] + (nn, nn))
    if return_full:
        return u_mean, u_cov, xfm, u, s, vt, nconf
    return u_mean, u_cov, xfm




import jax.numpy as np

from vbjax import make_region_mapping


def test_region_mapping():
    rm = np.r_[0, 0, 0, 1, 1, 2]
    to_surface, to_region = make_region_mapping(rm)
    x_region = np.r_[0.3, 0.5, -2.0]
    assert (to_region(to_surface(x_region)) == x_region).all()



"""
Functions for region mapping.

"""

import jax.numpy as np


def make_region_mapping(region_map):
    """Create functions, for a given region mapping, which generate
    the region average from a surface vector or expand a regional vector
    to a surface vector.
    """

    def to_surface(x_region):
        return x_region[region_map]

    vtx_count = np.bincount(region_map)
    def to_region(x_surface):
        x_region = np.zeros(vtx_count.size)
        x_region_sum = x_region.at[region_map].add(x_surface)
        return x_region_sum / vtx_count

    return to_surface, to_region



import jax.numpy as np


def shrinkage_zscore(x, x_hat, x_prior):
    """
    Computes the posterior shrinkage and z-score for 
    a parameter x.

    Parameters
    ----------
    x : array
        The true parameter.
    x_hat : array
        The posterior samples of the parameter.
    x_prior : array
        The prior standard deviation of the parameter.

    Returns
    -------
    shrinkage : array
        The posterior shrinkage.
    zscore : array
        The posterior z-score.

    """
    # compute posterior mean
    x_mean = x_hat.mean(0)

    # compute posterior standard deviation
    x_std = x_hat.std(0)

    # compute posterior shrinkage
    shrinkage = 1 - x_std / x_prior

    # compute posterior z-score
    zscore = (x_mean - x) / x_std

    return shrinkage, zscore



import itertools
import jax
import jax.numpy as np
from jax.example_libraries.optimizers import adam

from .layers import make_dense_layers
from .loops import make_ode


def embed_neural_flow(y, embed_dim=3, latent_dim=[256], iters=8000, step_size=0.01,
                      train_callback=None, oversample=10):
    """
    Embed a dataset using a neural flow.

    Parameters
    ----------
    y : array
        The dataset to embed, with shape (n_time, n_events).
    embed_dim : int
        The dimension of the embedding.
    latent_dim: list of int
        The dimension of the latent neural layers.
    iters : int
        The number of optimization iterations to run.


    Returns
    -------
    z : array
        The embedded dataset.

    """

    # embed y with gradients
    if embed_dim > 0:
        x = [y]
        for i in range( embed_dim - 1 ):
            x.append(np.gradient(x[-1], axis=1))
        x = np.array(x).transpose((2, 0, 1))
    else:
        x = y

    # define neural flow
    nn_p, nn_f = make_dense_layers(x.shape[1], latent_dim, init_scl=1e-3)
    x0 = x[0]
    dt = 1.0 / oversample
    step, loop = make_ode(dt, lambda x, p: nn_f(p, x))
    ts = np.r_[: x.shape[0]*oversample ]
    xp = loop(x0, ts, nn_p)

    # loss function
    def loss(nn_p):
        xp = loop(x0, ts, nn_p)
        return np.sum(np.square(x - xp[::oversample]))
    gloss = jax.jit( jax.grad( loss ) )

    # optimize
    opt = adam(step_size)
    opt_x = opt.init_fn(nn_p)
    trace = []
    trace_grads = []
    best_loss_so_far = np.inf
    best_so_far = nn_p
    for i in range(iters):
        p = opt.params_fn(opt_x)
        g = gloss(p)
        trace.append(loss(p))
        trace_grads.append(g)
        if trace[-1] < best_loss_so_far:
            best_loss_so_far = trace[-1]
            best_so_far = p
        if train_callback is not None:
            xp = loop(x0, ts, p)
            train_callback(i, p, g, trace[-1], x, xp)
        opt_x = opt.update_fn(i, g, opt_x)

    p = best_so_far
    xp = loop(x0, ts, p)

    # return embedded data
    return xp, trace, nn_f, p


def embed_polynomial(x, y=None, max_order=2):
    """
    Embed a dataset using a polynomial basis.

    Parameters
    ----------
    x : array
        The dataset to embed, with shape (n_dim, n_sample).
    y : array, optional
        The target data for a polynomial regression, (n_target, n_sample).
    max_order : int
        The maximum order of the polynomial basis.

    Returns
    -------
    basis : array
        The polynomial basis, with shape (n_basis, n_sample).
    coef : array, optional
        The coefficients of the polynomial regression, with shape (n_basis, n_target).
        This is only returned if `y` is provided.

    Notes
    -----
    The loss can be computed by 
    >>> np.square(coef.T @ basis - Y).sum()

    """

    n_dim = len(x)
    n_samp = len(x.T)
    basis = [np.ones((1, n_samp))]

    # construct columns of A matrix
    for order in range(1, max_order+1):
        all_idx = itertools.product(*([range(n_dim)] * order))
        x_idx = list({tuple(sorted(idx)): None for idx in all_idx}.keys())
        basis.append(np.prod(x[np.array(x_idx)], axis=1))
    basis = np.concatenate(basis, axis=0)

    if y is None:
        return basis

    # compute least square solution
    coef, *_ = np.linalg.lstsq(basis.T, y.T, rcond=None)

    return basis, coef


def embed_gradient(x, n_grads=2, axis=1):
    """
    Embed a dataset using repeated `np.gradient` calls to compute
    derivatives, usually for time series data.

    Parameters
    ----------
    x : array
        The dataset to embed, shape (n_dim, n_time, ...).
    n_grads : int
        The number of derivatives to compute.

    Returns
    -------
    basis : array
        The derivative basis, with shape (n_grads, n_dim, n_time, ...).

    """
    xs = [x]
    for i in range(n_grads):
        xs.append(np.gradient(xs[-1], axis=1))
    return np.array(xs)


def embed_autoregress(x, n_lag=3, lag=1):
    """
    Embed a dataset as an autoregressive process,
    usually for time series data.

    Parameters
    ----------
    x : array
        The dataset to embed, shape (n_dim, n_time, ...).
    n_lag : int
        The number of lags to use.
    lag : int, default 1
        The lag to use.

    Returns
    -------
    basis : array
        The autoregressive basis, with shape (n_lag, n_dim, n_time - lag*n_lag, ...)

    """
    xs = np.array([x[:, i*lag:(-n_lag+i)*lag] for i in range(n_lag)])
    return xs



from . import test_loops, test_regmap, test_shtlc, test_coupling

test_mods = [test_loops, test_regmap, test_shtlc, test_coupling]

for mod in test_mods:
    for key in dir(mod):
        if key.startswith('test_'):
            test_fn = getattr(mod, key)
            test_fn()

print('vbjax tests passed')


import numpy as np
import vbjax


def test_embed_polynomial():
    X = x, y, z = np.random.randn(3, 100)
    basis = vbjax.embed.embed_polynomial(X, max_order=2)
    assert basis.shape == (10, 100)

    Y = np.c_[x - x**2, z + x*y].T
    basis, coef = vbjax.embed.embed_polynomial(X, Y, max_order=2)
    assert basis.shape == (10, 100)
    assert coef.shape == (10,2)
    assert np.square(coef.T @ basis - Y).sum() < 1e-6


if __name__ == '__main__':
    test_embed_polynomial()
